use evalexpr::{
	Value,

	EvalexprError
};

use crate::util;

pub type EvalResult = Result<Value, EvalexprError>;


//	Mathematics
pub fn cosine(arg: &Value) -> EvalResult {
	Ok(
		if		let Value::Float(float)	= arg { float.clone() }
		else if	let Value::Int(int)		= arg { int.clone() as f64 }
		else { return Err(EvalexprError::expected_number(arg.clone())) }
		.cos().into()
	)
}

pub fn fix(arg: &Value) -> EvalResult {
	if let Value::Tuple(args) = arg {
		let len = args.len();
		if len == 2 {
			let value =
				if let Value::Float(float) = args[0] { float.clone() }
				else { return Err(EvalexprError::expected_float(args[0].clone())); };

			let operand = 10u64.pow(
				if let Value::Int(int) = args[1] { int.clone() }
				else { return Err(EvalexprError::expected_int(args[1].clone())); }
				as u32
			) as f64;

			Ok( ((value * operand).round() / operand).into() )
		} else { Err(EvalexprError::wrong_function_argument_amount(len, 2)) }
	} else { Err(EvalexprError::wrong_function_argument_amount(1, 2)) }
}

pub fn logarithm(arg: &Value) -> EvalResult {
	let value: f64;
	let base: Option<f64>;
	match arg {
		Value::Tuple(tuple)
			=>	{
				let len = tuple.len();
				if len != 2 { return Err(EvalexprError::wrong_function_argument_amount_range(len, 1..=2)); }

				let i_value = tuple.get(0).unwrap();
				if		let Value::Float(float)	= i_value { value = float.clone(); }
				else if	let Value::Int(int)		= i_value { value = int.clone() as f64; }
				else { return Err(EvalexprError::expected_number(i_value.clone())); }

				let i_base = tuple.get(1).unwrap();
				if		let Value::Float(float)	= i_value { base = Some(float.clone()); }
				else if	let Value::Int(int)		= i_value { base = Some(int.clone() as f64); }
				else { return Err(EvalexprError::expected_number(i_base.clone())); }
			},
		Value::Float(float)
			=>	{
				value = float.clone();
				base = None;
			},
		Value::Int(int)
			=>	{
				value = int.clone() as f64;
				base = None;
			}
		_	=>	return Err(EvalexprError::CustomMessage("Expected numbers.".to_owned()))
	}

	if let Some(base) = base { Ok(value.log(base).into()) }
	else { Ok(value.ln().into()) }
}

pub fn sine(arg: &Value) -> EvalResult {
	Ok(
		if		let Value::Float(float)	= arg { float.clone() }
		else if	let Value::Int(int)		= arg { int.clone() as f64 }
		else { return Err(EvalexprError::expected_number(arg.clone())) }
		.sin().into()
	)
}

pub fn square_root(arg: &Value) -> EvalResult {
	Ok(
		if		let Value::Float(float)	= arg { float.clone() }
		else if	let Value::Int(int)		= arg { int.clone() as f64 }
		else { return Err(EvalexprError::expected_number(arg.clone())) }
		.sqrt().into()
	)
}

pub fn tangent(arg: &Value) -> EvalResult {
	Ok(
		if		let Value::Float(float)	= arg { float.clone() }
		else if	let Value::Int(int)		= arg { int.clone() as f64 }
		else { return Err(EvalexprError::expected_number(arg.clone())) }
		.tan().into()
	)
}


//	Data Science
pub fn average(arg: &Value) -> EvalResult {
	if let Value::Tuple(args) = arg {
		let len = args.len() as f64;
		let mut total = 0f64;

		for arg in args {
			total +=
				if		let Value::Float(float) = arg { float.clone() }
				else if	let Value::Int(int)		= arg { int.clone() as f64 }
				else { todo!() };
		}

		Ok( (total / len).into() )
	} else { todo!() }
}

//	Radix conversion
pub fn binary(arg: &Value) -> EvalResult {
	if let Value::Int(int) = arg { Ok( format!("0b{:b}", int).into() ) }
	else if let Value::String(string) = arg { util::parse_radix("0b", 2, string) }
	else { Err(EvalexprError::expected_number_or_string(arg.clone())) }
}

pub fn hexadecimal(arg: &Value) -> EvalResult {
	if let Value::Int(int) = arg { Ok( format!("0x{:X}", int).into() ) }
	else if let Value::String(string) = arg { util::parse_radix("0x", 16, string) }
	else { Err(EvalexprError::expected_number_or_string(arg.clone())) }
}

pub fn octal(arg: &Value) -> EvalResult {
	if let Value::Int(int) = arg { Ok( format!("0o{:#o}", int).into() ) }
	else if let Value::String(string) = arg { util::parse_radix("0o", 8, string) }
	else { Err(EvalexprError::expected_number_or_string(arg.clone())) }
}

