use std::{
	env,
	io::{
		stdin,
		stdout,

		IsTerminal,
		Read,
		Write
	},
	process::exit
};

use evalexpr::{
	eval_with_context_mut,

	ContextWithMutableVariables,
	EvalexprError,
	HashMapContext,
	Value
};
use pico_args::Arguments;
use rustyline::DefaultEditor;
use termion::{
	color,
	style
};

mod context;
mod flag;
mod util;

pub const VERSION: &str = env!("CARGO_PKG_VERSION");

fn main() {
	let mut args = Arguments::from_env();

	//	handle breaking flags
	if args.contains(flag::HELP) {
		help_text();
		return;
	}
	if args.contains(flag::VERSION) {
		version_text();
		return;
	}

	let mut context = context::build(&mut args);
	let is_terminal = stdin().is_terminal() && stdout().is_terminal();
	let quiet = args.contains(flag::QUIET) || !is_terminal;

	//	collect args and evaluate if present
	let expressions = args.finish();
	if expressions.len() > 0 {
		for expression in expressions {
			let expression: String = expression.to_string_lossy().into();
			eval(&expression, &mut context, quiet);
		}
	} else {
		if !is_terminal {
			let mut buffer = String::with_capacity(0);
			if let Ok(_) = stdin().read_to_string(&mut buffer) {
				for expression in buffer.lines() {
					eval(&expression.to_string(), &mut context, quiet);
				}
			}
		} else if let Ok(mut rl) = DefaultEditor::new() {
			//	enter interactive mode if no args are given
			version_text();
			println!("{}Interactive Mode{}", style::Faint, style::Reset);
			loop {
				if let Ok(line) = rl.readline("> ") {
					match line.as_str() {
						""		|
						"exit"	=>	break,
						_		=>	eval(&line, &mut context, quiet)
					}
				}
				reset();
			}
		}
	}
}

fn eval(expression: &str, context: &mut HashMapContext, quiet: bool) {
	let result = eval_with_context_mut(expression, context);

	if quiet {
		if let Ok(result) = result { println!("{result}") }
		else { exit(1) }
	} else { format(expression, result) }
}

fn format(expression: &str, result: Result<Value, EvalexprError>) {
	if let Ok(result) = result {
		if !result.is_empty() {
			let delimiter =
				match result {
					Value::Boolean(_)	=>	"is",
					Value::String(_)	=>	"=>",
					_					=>	"="
				};
			println!(
				"{faint}{italic}{expression}{reset} {delimiter} {bold}{result}",
				bold	=	style::Bold,
				faint	=	style::Faint,
				italic	=	style::Italic,
				reset	=	style::Reset
			);
		} else {
			println!(
				"{green}✓ {bold}{expression}",
				bold = style::Bold,
				green = color::Fg(color::Green)
			);
		}
	} else {
		println!(
			"{red}✕ {bold}{expression}",
			bold = style::Bold,
			red = color::Fg(color::Red)
		);
	}
}

fn reset() {
	print!("{}{}", style::Reset, color::Fg(color::Reset));
	stdout().flush().unwrap();
}

fn version_text() {
	println!("qm v{VERSION}");
}

fn help_text() {
	version_text();
	println!(
"Valerie Wolfe <sleeplessval@gmail.com>
A mathematical expression evaluator written in Rust.

usage:
   qm [EXPRESSION]..."
	);
}
